// -----------------------------------------------------------------------------
// xMatters Integration for HP Operations Manager for Linux
//
// Main JavaScript file for the integration
// -----------------------------------------------------------------------------

importPackage(java.lang);
importPackage(java.util);
importPackage(java.util.regex);
importPackage(java.io);
importPackage(java.sql);
importClass(Packages.com.alarmpoint.integration.hpomu.HPOMUInteractionWrapper);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.alarmpoint.integrationagent.util.EventDeduplicator);
importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);

var xStream = new XStream(new PureJavaReflectionProvider())

var INT_VERSION = "hpoml-1-0-1";

// Core Javascript files provided by the IA
load("integrationservices/" + INT_VERSION + "/lib/javascript/core/baseclass.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/core/logger.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/core/util.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/webservices/wsutil.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/webservices/soapfault.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/xmatters/xmattersws.js");

load("integrationservices/" + INT_VERSION + "/hpoml-config.js");

var exceptionString = "JavaException: ";
var notePrefix = "[xMatters] - "; // The Prefix which is added to the note comments e.g. "[xMatters] "
var calloutUser = "xMatters"; // Username associated with callout annotations

/**
 * This is the main method for handling APXML messages sent to the 
 * Integration Service by APClient.  The messages may be requests to perform 
 * local activity, or they may be requests to make submissions to AlarmPoint.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * via the Integration Service's outbound queues.
 */
function apia_input(apxml) {
    var webService = new xMattersWS();
    log.info("Received an injection with tokens " + webService.convertTokensToString(apxml));
    
    // If backward compatibility required, set recipient based on message group
    if (useMessageGroupAsRecipient)
    {
      set_recipient_to_message_group(apxml);
    }

    // Extract any custom variables from the instruction text. Only needed if
    // custom message attributes are not available.
    extract_variables_from_instructions(apxml);

    // Add custom message attributes. Note that these override anything that may
    // have been set before in set_recipient_to_message_group() or
    // extract_variables_from_instructions().
    split_custom_variables(apxml);

    // injected message fields
    var recip = apxml.getValue("recipients");
    var incidentId = apxml.getValue("incident_id");
    
    // Annotate the original OMU message with the fact that it has been sent to AlarmPoint
    if (recip != null && recip != "")
    {
      annotateOnSubmission("Notifying xMatters for recipient(s) " + recip, incidentId);
    } else
    {
      annotateOnSubmission("Notifying xMatters for subscribers", incidentId);
    }

    // Enrich the APXML with the layout group information
    retrieveLayoutGroup(apxml);

    // Forward enriched APXML to AlarmPoint, subject to deduplication
	try
	{
    if (webService.submitAPXML(apxml, DELETE_EXISTING_EVENTS))
    {
        // If submitAPXML() actually sent the message and didn't deduplicate it, add an annotation to the IA logs.
        log.info(notePrefix + "Notification will be sent to [" + apxml.getValue("recipients") + "].");
    }
	}
	catch (ex)
	{
		ServiceAPI.getLogger().error("an exception was encountered while applying the deduplication filter.  Exception is: " + ex.toString());
	}
}

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to 
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint 
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a 
 * Web Service response, depending on the mechanism used to submit the APXML.
 */
function apia_response(apxml) 
{
    var result = null;
    var webService = new xMattersWS();
    log.info( "apia_response() received APXML from xMatters; content is: " + webService.convertTokensToString(apxml));

    var method = apxml.getMethod();

    log.debug("Method =[" + method + "]");
  
    if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
    {
        result = handleOK(apxml);
    }
    else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
    {
        result = handleERROR(apxml);
    }
    else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
    {
        result = handleSend(apxml);
    }
    else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
    {
        result = handleRequest(apxml);
    }
    else
    {
        throw new IllegalArgumentException("Unrecognized APXML method: " + method);
    }

    log.debug("Exit - apia_response");
    return result;
}

/**
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * <p>
 * Log the acknowledgement using the ping Integration Service's logger.
 */
function handleOK(apxml)
{
    var id = apxml.getTransactionId();
    log.debug("Received 'OK' response from xMatters for APXML message with ID: " + id );
}

/**
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * <p>
 * Log the error using the ping Integration Service's logger.
 */
function handleERROR(apxml) 
{
    var id = apxml.getTransactionId();
    var code = apxml.getValue(APXMLMessage.ERROR_CODE);
    log.error("Submission of transaction " + id + " failed with error code " + code);
}

/**
 * Requests are forwarded to the Send handler.
 * <p>
 * Redirects to the handleSend as messages may come from ESM or ESR to the IA.
 */
function handleRequest(apxml)
{
    log.debug( "\tEnter - handleRequest");
    return handleSend(apxml);
}

/**
 * The main routine for handling Requests from AlarmPoint
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleSend(apxml)
{
    log.debug( "\t\tEnter - handleSend");
    var result = null;

    // The presence of the message token means this message is a callout annotation.
    var message = apxml.getValue("message");
    var incidentId = apxml.getToken("incident_id").getValue();

    if ( message != null && incidentId != null)
    {
        result = handleCalloutAnnotate(incidentId, message);
    }
    else
    {
        result = handleApsResponses( apxml );
    }

    log.debug( "\t\tExit - handleSend");
    return result;
}

/**
 * -----------------------------------------------------------------------------
 * handleApsResponses
 *
 * Handling of delivery notifications and user responses from xMatters.
 * Decide what to do with the APXML message that has been received.
 * -----------------------------------------------------------------------------
 */
function handleApsResponses(apxml)
{
    log.debug( "\t\t\tEnter - handleApsResponses");

    // Initialize xMattersWS interface building a reply to xMatters.
    var xmWebService = new xMattersWS();
    var responseToXM = xmWebService.initializeResponse(apxml);
    var responseTokens = getTokens(apxml);

    try
    {
        // deserialize the additional tokens
        log.debug( "\t\t\thandleApsResponses - Token keys from xMatters: " + responseTokens);

        var responseAction = responseTokens.get("responseAction");
        var annotation = responseTokens.get("annotation");

        // See if the APXML message is invalid, or if we don't have anything to do
        // in response to it.
        if (!checkAPXMLTokens(responseTokens))
        {
            // Nothing to do in response to this message, so just return OK
            responseToXM.setToken(xmWebService.TOKEN_NAME, "OK", APXMLToken.Type.STRING);
        }
        else
        {
            if (responseTokens.get("responseEvent") == "SUCCESSFUL_DELIVERY"
                || responseTokens.get("responseEvent") == "DELIVERY_FAILURE")
            {
                log.debug("Received delivery notification [" + responseTokens.get("responseEvent") + "]");
                // Create an annotation for the successful or failed delivery of a notification.
                // Set the responseAction token to Annotate to pass control to the Annotate handler.
                responseAction = "Annotate";
                annotation = responseTokens.get("responseEvent").toProperCase() + " to " + responseTokens.get("responder");
            }

            handleResponseAction(responseAction, annotation, responseTokens, responseToXM);

            responseToXM.setToken(xmWebService.TOKEN_NAME, "OK", APXMLToken.Type.STRING);
        }
    }
    catch (e)
    {
        var errMsg = "Caught exception processing [" + responseTokens.get("responseEvent") + "|" + responseAction + "]: " + e;
        log.error(errMsg);
        responseToXM.setToken(xmWebService.TOKEN_NAME_EXCEPTION, errMsg.replace(exceptionString, ""), APXMLToken.Type.STRING);
    }
    
    log.debug( "\t\t\tExit - handleApsResponses");
    return responseToXM;
}

/**
 * -----------------------------------------------------------------------------
 * handleResponseAction
 *
 * Dispatch the user response to the appropriate handler function. The caller
 * may have over-ridden the responseAction and annotation tokens from the APXML
 * message, but the other tokens will be untouched.
 * -----------------------------------------------------------------------------
 */
function handleResponseAction(responseAction, annotation, responseTokens, responseToXM)
{
    log.debug("\t\t\tEnter - handleResponseAction. Handling responseAction [" + responseAction + "] for incidentID [" + responseTokens.get("incidentId") + "]");

    var incidentId = responseTokens.get("incidentId");
    var responder = responseTokens.get("responder");
    var responderId = responseTokens.get("responderId");
    var sendAnnotations = "true".equalsIgnoreCase(responseTokens.get("sendAnnotations"));
    var isRepliedFromXM = true;

    
    log.debug("Loging into HPOM with user " + omu_user);
    var wrapper = new HPOMUInteractionWrapper(omu_executable, sql_executable, omu_user, getPassword());

    // Validate to make sure the credentials supplied above (omu_user, omu_password, omu_executable) are correct
    wrapper.validateCredentials();

    // extract message data
    var incidentId = responseTokens.get("incident_id");
    var requestText = responseTokens.get("request_text");
    var messageText = responseTokens.get("message_text");
    var requestValue = responseTokens.get("request_value");
    var message = responseTokens.get("message");
    
    if (annotation != null)
    {
        annotation = notePrefix + annotation;
    }

    switch (String(responseAction).toLowerCase())
    {
        case "acknowledge":
            if (responder)
            {
              wrapper.acknowledgeMessage(incidentId);
              wrapper.addAnnotation(incidentId, notePrefix + " Message Acknowledged by " + responder);
              ServiceAPI.getLogger().info("Message Acknowledged by " + responder);
            }
            else
            {
              ServiceAPI.getLogger().error("No responder set on acknowledgement");
            }
            break;
        case "own":
            if (responderId)
            {
              wrapper.ownMessage(incidentId);
              wrapper.addAnnotation(incidentId, notePrefix + " Message Owned by " + responder);
              ServiceAPI.getLogger().info("Message Owned by " + responder);
            }
            else
            {
              ServiceAPI.getLogger().error("Request to claim ownership provided no owner ");
            }
            break;
        case "changeseverity":
            if (requestValue)
            {
              wrapper.updateMessage(incidentId, null, requestValue);
              wrapper.addAnnotation(incidentId, notePrefix + " Severity Changed to " + requestValue + " by " + responder);
              ServiceAPI.getLogger().info("Severity Changed to " + requestValue);
            }
            else
            {
              ServiceAPI.getLogger().error("Request to change severity provided no severity value");
            }
            break;
        case "ignore":
			wrapper.addAnnotation(incidentId, notePrefix + " Ignored by " + responder);
			ServiceAPI.getLogger().info("Ignored by " + responder);
            break;
        case "annotate":
            wrapper.addAnnotation(incidentId, annotation);
            break;

        default:
            throw "Unknown response Action: " + responseAction;
            break;
    }

    log.debug("\t\t\tExit - handleResponseAction");
}

/*
 * checkAPXMLTokens
 *
 * Check that the received response tokens are valid and see if this message needs to be acted on. 
 * Throws an exception if message is invalid.
 * Returns true if the message needs to be processed.
 */
function checkAPXMLTokens(responseTokens)
{
    var result = true;

    switch (String(responseTokens.get("responseEvent")))
    {
        case "RECEIVED_RESPONSE":
        case "STANDALONE_ANNOTATION":
            if(responseTokens.get("responseAction") == null
                || responseTokens.get("responseAction") == ""
                || responseTokens.get("responseAction") == undefined)
                {
                throw "responseAction not set for a RECEIVED_RESPONSE responseEvent.  responseAction required check AlarmPoint scripting to ensure this token is set";
            }
            break;

        case "DEVICE_DELIVERED":
            // Log voice device delivery, but do not annotate to the MS, a callout annotation will be used instead.
            // So just return.
            log.debug(notePrefix + responseTokens.get("responseEvent") + " to " + responseTokens.get("responder"));
            result = false;
            break;

        case "SUCCESSFUL_DELIVERY":
        case "DELIVERY_FAILURE":
            if (!ANNOTATE_DELIVERY)
            {
                log.warn("DELIVERY_ANNOTATIONS are suppressed the following message will not be annotated to event: ["
                    + responseTokens.get("incidentId") + "] " + responseTokens.get("annotation"));
                result = false;
            }
            break;

        default:
            throw "Unknown responseEvent.  Expected: RECEIVED_RESPONSE||SUCCESSFUL_DELIVERY||DELIVERY_FAILURE||DEVICE_DELIVERED. Received: " 
            + responseTokens.get("responseEvent");
            break;
    }
   
    return result;
}

// -----------------------------------------------------------------------------
// Callout annotation processing
//
// If enabled, these result in annotations (notes) being added to the event.
// -----------------------------------------------------------------------------
function handleCalloutAnnotate(incidentId, messageText)
{
    log.debug("Enter - handleCalloutAnnotate - incidentId [" + incidentId + "], messageText [" + messageText + "]");

    if(!ANNOTATE_DELIVERY)
    {
        log.warn("DELIVERY_ANNOTATIONS are suppressed the following message will not be annotated to event: [" + incidentId + "] " + messageText);
        return;
    }

    log.debug( "Starting call to management system for callout annotate for [" + incidentId + "]" );

    
    log.debug("Exit - handleCalloutAnnotate");
}

// -----------------------------------------------------------------------------
// Utility functions
// -----------------------------------------------------------------------------

/**
 * Retrieves the additional tokens from the APXML
 * @param apxml XML message to process
 * @return additional tokens
 * @throws if the additional tokens are missing from the XML.
 */
function getTokens(apxml)
{
    if (apxml.getValue("additionaltokens") != null && !"".equalsIgnoreCase(apxml.getValue("additionaltokens")))
    {
        return xStream.fromXML(apxml.getValue("additionaltokens")); // LinkedHashMap
    }
    throw new Exception("additionaltokens is missing from apxml");
}




/**
 * Retrieve the layout groups from the database. SQL scripts are located in
 * /etc/opt/OV/share/conf/OpC/mgmt_sv/reports/C/ and run using the call_sqlplus
 * command provided. This method will return the data in a comma separated list
 * @param apxml instance of APXML that the layout group will be appended to
 */
function retrieveLayoutGroup(apxml)
{
  var sourceNode = apxml.getValue("source_node");
  ServiceAPI.getLogger().debug("About to lookup layout groups for " + sourceNode);
  var results = executeCommand(sql_executable, ["layouts",sourceNode]);

  results = results.replace("\n", ",");
  apxml.setToken("layout_group", results);
  apxml.setToken("layout_group-list-item-delimiter", ",");
}

/**
 * Execute a shell command
 * @param cmd name of the command to execute
 * @param arguments array of arguments that the command expects
 * @return the console output from the command call
 */
function executeCommand(cmd, arguments)
{
  try
  {
    // variable used by runCommand
    var options = {
     args: arguments,
     output: ''
    };

    // execute script
    runCommand(cmd, options);
    if (ServiceAPI.getLogger().isDebugEnabled())
    {
      ServiceAPI.getLogger().debug("runCommand options.output: " + options.output);
    }
    return options.output; 
  }
  catch (e)
  {
    ServiceAPI.getLogger().error("There was an error executing the command. Exception: " + e.toString()); 
  }
}

/**
 * DESCRIPTION
 * This procedure will enrich the injected OMU variables and pass them to
 * AlarmPoint.
 *
 * GIVEN INPUTS
 * apxml - The injected OMU message with the OMU variables.
 */
function split_custom_variables(apxml)
{
  var cma = apxml.getValue("cma");
  if (cma)
  {
    var kv_pairs = cma.split(";;");
    var kv_count = kv_pairs.length;

    // For each line of the custom text grab the key and value
    for (var x = 0; x < kv_count; x++)
    {
      var curr_kv_pair = kv_pairs[x];
      var key_value = curr_kv_pair.split("=", 2);
      if (key_value.length == 2 && key_value[0].length() > 0 && key_value[1].length() > 0)
      {
        var key = key_value[0];
        var value = key_value[1];

        // remove all spaces and change it to use underscores
        key = key.replaceAll(" ", "_");

        // Add the pair to the AP XML Message to be injected to AlarmPoint
        apxml.setToken(key, value);

        // Concatenate the tokens added to the event to the log for debugging purposes
        if (ServiceAPI.getLogger().isDebugEnabled())
        {
          ServiceAPI.getLogger().debug("Added custom parameter: " + key + " = " + value);
        }
      }
    }
  }
}


/**
 * DESCRIPTION
 * This procedure will enrich the injected OMU variables and pass them to
 * AlarmPoint by pulling them from the instruction text. Note that the
 * Custom Message Attributes (see split_custom_variables()) is the preferred
 * way to accomplish this, but it is not available on every message template
 * type. So this gives us a backup mechanism.
 *
 * GIVEN INPUTS
 * apxml - The injected OMU message with the OMU variables.
 */

function extract_variables_from_instructions(apxml)
{
  var instruction = apxml.getValue("instruction");
  if (instruction)
  {
    ServiceAPI.getLogger().debug("Instruction in APXML is: " + instruction );

    var variable;
    var value;
    var adjusted = false;

    // Configure regex matching for strings with the following format
    // AP-<variable>:<value>
    var regexp = "^AP-(\\w+):(.+)$";
    var pattern = Pattern.compile(regexp, Pattern.MULTILINE);
    var matcher = pattern.matcher(instruction);

    var EndofCustomVars = 0;

    // For each line of the custom text grab the variable and value
    while (matcher.find())
    {
      variable = matcher.group(1);
      value = matcher.group(2);
      EndofCustomVars = matcher.end() + 1;

      // Add the pair to the AP XML Message to be injected to AlarmPoint
      apxml.setToken(variable, value);

      // Concatenate the added variables to the log for debugging purposes
      if (ServiceAPI.getLogger().isDebugEnabled())
      {
        ServiceAPI.getLogger().debug("Added instruction custom parameter: " + variable + " = " + value);
      }
      adjusted = true;
    }

    if (adjusted)
    {
      if (instruction.length > EndofCustomVars)
      {
        instruction = instruction.substring(EndofCustomVars);
        if (ServiceAPI.getLogger().isDebugEnabled())
        {
          ServiceAPI.getLogger().debug("Instruction text Adjusted: " + instruction);
        }
      } else
      {
        instruction = "";
        if (ServiceAPI.getLogger().isDebugEnabled())
        {
          ServiceAPI.getLogger().debug("Instruction text Consumed");
        }
      }
      // set the token back in the message
      apxml.setToken("instruction", value);
    }
  }
}

/**
 * DESCRIPTION
 * This procedure sets the recipient of the event to the name of the
 * OVO Message group, optionally with a prefix added. This mechanism is
 * deprecated in favour of setting the recipients on the OVO message
 * template using custom message attributes (see split_custom_variables()
 * method above).
 *
 * GIVEN INPUTS
 * apxml - The injected OMU message with the OMU variables.
 */
function set_recipient_to_message_group(apxml)
{
  var messageGroup = apxml.getValue("message_group");
  if (messageGroup)
  {
    var recipient;
    if (messageGroupRecipientPrefix && messageGroupRecipientPrefix.length > 0)
    {
      recipient = messageGroupRecipientPrefix + messageGroup;
    } else
    {
      recipient = messageGroup;
    }
    // Set the recipients token on the event to the message group with any prefix
    apxml.setToken("recipients", recipient);

    // Note the setting of the recipients token in the log message
    if (ServiceAPI.getLogger().isDebugEnabled())
    {
      ServiceAPI.getLogger().debug(" Added recipients: " + recipients + " based on message group: " + messageGroup);
    }
  } else
  {
    // throwing exception to add message and still break the flow
    throw new RuntimeException("Agent Client: hpoml-1-0-1 - Could not set recipient as no message group specified!  Error: message_group not set.");
  }
}

/**
 * DESCRIPTION
 * This procedure will annotate the original OMU message with the
 * information that the message has been sent into AlarmPoint.
 *
 * GIVEN INPUTS
 * annotate - The annotation to add to the OMU message.
 * incidentId - the unique id for the incident
 */
function annotateOnSubmission(annotation, incidentId)
{
  var wrapper = new HPOMUInteractionWrapper(omu_executable, sql_executable, omu_user, getPassword());

  // Validate to make sure the credentials supplied above (omu_user, omu_password, omu_executable) are correct
  wrapper.validateCredentials();
  wrapper.addAnnotation(incidentId, annotation);
}

/**
 * Retrieves the password from the encrypted password file
 */
function getPassword()
{ 
  var encryptionUtils = new EncryptionUtils(); 
  var passwordFile = new File(PASSWORD_FILE); 
  var password = encryptionUtils.decrypt(passwordFile); 
  return password;
}
